#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import fnmatch
import re
import sys

import json
import boto3

from .cfn_updater import CfnUpdater
from .replace_references import replace_references


def split_resource_name(resource_name):
    m = re.match(r'^(?P<base>.*)v(?P<version>[0-9]+)$', resource_name)

    if m is not None:
        base = m.group('base')
        version = int(m.group('version'))
    else:
        base = resource_name
        version = 0

    return base, version


def make_new_resource_name(resource_name):
    base, version = split_resource_name(resource_name)
    return '{}v{}'.format(base, version + 1)


class AMIUpdater(CfnUpdater):
    """
    Updates the AMI name of Custom::AMI resources to the latest version. 
    With an ami-name-pattern of `amzn-ami-*ecs-optimized` it will update
    the following resource definition from:

\b
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.a-amazon-ecs-optimized
           Owners:
             - amazon

    to:

\b
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.l-amazon-ecs-optimized
           Owners:
             - amazon


    By specifying --add-new-version, a new Custom::AMI will be added
    to the template with a new name. A suffix `v<version>` is appended
    to create the new resource. Any reference to the original Custom::AMI
    resource is replaced. It will change:

\b
      CustomAMI:
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.a-amazon-ecs-optimized
           Owners:
             - amazon
      CustomAMIv2:`
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.b-amazon-ecs-optimized
           Owners:
             - amazon
      Instance:
         Type: AWS::EC2::Instance
         Properties:
            ImageId: !Ref CustomAMIv2

    to:

\b
      CustomAMI:
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.a-amazon-ecs-optimized
           Owners:
             - amazon
      CustomAMIv2:
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.b-amazon-ecs-optimized
           Owners:
             - amazon
      CustomAMIv3:
         Type: Custom::AMI
         Properties:
           Filters:
             name: amzn-ami-2017.09.l-amazon-ecs-optimized
           Owners:
             - amazon
      Instance:
         Type: AWS::EC2::Instance
         Properties:
            ImageId: !Ref CustomAMIv3
    """

    def __init__(self):
        super(AMIUpdater, self).__init__()
        self._ami_name_pattern = None
        self.add_new_version = False
        self.latest_ami_name_pattern = None

    @property
    def ami_name_pattern(self):
        return self._ami_name_pattern

    @ami_name_pattern.setter
    def ami_name_pattern(self, name):
        self._ami_name_pattern = name

    @staticmethod
    def is_ami_definition(resource):
        return resource.get('Type', '') == 'Custom::AMI'

    def create_describe_image_request(self, ami):
        # copy the filters values, except for name and state.
        properties = ami.get('Properties', {})
        filters = [{'Name': 'name', 'Values': [self.ami_name_pattern]},
                   {'Name': 'state', 'Values': ['available']}]
        for k, v in properties.get('Filters', {}).items():
            if k not in ['name', 'state']:
                filters.append({'Name': k, 'Values': v if isinstance(v, list) else [v]})

        # copy the rest of the arguments
        result = {n: (properties.get(n) if isinstance(properties.get(n), list) else [properties.get(n)])
                  for n in filter(lambda n: properties.get(n) is not None, ['Owners', 'ImageIds', 'ExecutableUsers'])}
        result['Filters'] = filters
        return result

    def load_latest_ami_name_pattern(self, resource):
        kwargs = self.create_describe_image_request(resource)
        response = boto3.client('ec2').describe_images(**kwargs)
        images = sorted(response['Images'], key=lambda i: i['CreationDate'])
        self.latest_ami_name_pattern = images[-1]['Name'] if len(images) > 0 else None
        if len(images) > 0:
            sys.stderr.write(
                'INFO: using {} matching {}, created on {}\n'.format(self.latest_ami_name_pattern,
                                                                     self.ami_name_pattern,
                                                                     images[-1]['CreationDate']))

    def is_ami_name_pattern_match(self, name):
        regex = re.compile(fnmatch.translate(self.ami_name_pattern))
        return regex.match(name)

    def all_custom_ami_resources(self):
        r = self.resources
        return filter(lambda n: self.is_ami_definition(r[n]) and self.is_name_filter_match(r[n]), r)

    def is_name_filter_match(self, ami):
        if 'Properties' in ami and 'Filters' in ami['Properties'] and 'name' in ami['Properties']['Filters']:
            return self.is_ami_name_pattern_match(ami['Properties']['Filters']['name'])
        return False

    def custom_ami_resources_partitions(self):
        partitions = {}
        ami_resources = self.all_custom_ami_resources()
        for resource_name in ami_resources:
            base, _ = split_resource_name(resource_name)
            if base not in partitions:
                partitions[base] = {}
            partitions[base][resource_name] = self.resources[resource_name]

        for base in partitions:
            yield partitions[base]

    @staticmethod
    def latest_custom_ami_resource(ami_resources):
        result = None
        highest_version = -1
        for resource_name in ami_resources:
            _, version = split_resource_name(resource_name)
            if version > highest_version:
                result = resource_name
                highest_version = version

        return result

    def ami_requires_update(self, ami):
        name = ami['Properties']['Filters']['name']
        return self.latest_ami_name_pattern is not None and name != self.latest_ami_name_pattern

    def update_ami(self, resource_name, ami):
        if self.ami_requires_update(ami):
            ami['Properties']['Filters']['name'] = self.latest_ami_name_pattern
            sys.stderr.write(
                'INFO: updating AMI definition "{}" name filter to {} in {}\n'.format(
                    resource_name, self.latest_ami_name_pattern, self.filename))
            self.dirty = True
        else:
            if self.verbose:
                sys.stderr.write(
                    'INFO: AMI definition "{}" name already up to date in {}\n'.format(
                        resource_name, self.filename))

    def update_inplace(self):
        """
        updates the name filter of AMI resource definitions in `self.template`.
        """
        for resource_name in self.all_custom_ami_resources():
            ami = self.resources[resource_name]
            self.load_latest_ami_name_pattern(ami)
            self.update_ami(resource_name, ami)

    def add_new_ami_resource(self):
        """
        add a new version of the AMI resource definition in `self.template`.
        """
        for ami_resources in self.custom_ami_resources_partitions():
            resource_name = self.latest_custom_ami_resource(ami_resources)
            if resource_name is not None:
                ami = json.loads(json.dumps(self.resources[resource_name]))
                self.load_latest_ami_name_pattern(ami)
                if self.ami_requires_update(ami):
                    new_resource_name = make_new_resource_name(resource_name)
                    self.resources[new_resource_name] = ami
                    self.update_ami(new_resource_name, ami)
                    for old_resource_name in ami_resources:
                        replace_references(self.template, old_resource_name, new_resource_name)

    def update_template(self):
        if self.add_new_version:
            self.add_new_ami_resource()
        else:
            self.update_inplace()

    def main(self, ami_name_pattern, dry_run, verbose, add_new_version, path):
        self.dry_run = dry_run
        self.verbose = verbose
        self.ami_name_pattern = ami_name_pattern
        self.add_new_version = add_new_version
        self.load_latest_ami_name_pattern({})
        if self.latest_ami_name_pattern is None:
            sys.stderr.write('ERROR: image name {} does not resolve to an active AMI \n'.format(self.ami_name_pattern))
            sys.exit(1)

        self.update(path)



import re
from aws_cfn_update.cfn_updater import Ref

def replace_references(template, old_reference, new_reference):
    """
    replaces CloudFormation references { "Ref": old_reference } with { "Ref": new_reference } in `template`.
    """
    if isinstance(template, dict):
        for name, value in template.items():
            if name == 'Ref' and value == old_reference:
                template['Ref'] = new_reference
            elif isinstance(value, Ref) and value.reference == old_reference:
                value.reference = new_reference
            else:
                replace_references(template[name], old_reference, new_reference)
    elif isinstance(template, list):
        for i, value in enumerate(template):
            replace_references(value, old_reference, new_reference)
    else:
        pass # a primitive, no recursive required.



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
import cfn_flip
import os.path
import json
import collections
from ruamel.yaml import YAML


class CfnUpdater(object):
    """
    base class for a CloudFormation  update. To implement a specific updater: 

    - inherit this class
    - override the method update_template()
    - call the method update(path)

    it will read the template from files with extension .yaml, .yml and .json into 
    the property `self.template`.

    If the property `self.dirty` is set to True, the template will be written 
    back to the originating file.

    Please note that formatting and comments may be lost, when using this 
    updater.
    """

    def __init__(self):
        self.basename = None
        self.template = False
        self.template = None
        self.template_format = None
        self.dirty = False
        self.dry_run = False
        self.verbose = False
        self._filename = None

    @property
    def filename(self):
        """
        current template filename
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        """
        requires `filename` ends with `.json`, `.yaml` or `.yml`.
        sets `basename`, `template_format` and `_filename` accordingly.
        clears `dirty` and `templae`
        """
        self._filename = filename
        parts = os.path.splitext(os.path.basename(filename))
        self.basename = parts[0]
        self.template_format = parts[1]
        self.template = None
        self.dirty = False
        if self.template_format not in ('.json', '.yml', '.yaml'):
            raise ValueError('%s has no .json, .yaml or .yml extension.' % filename)

    def load(self):
        """
        loads `template` from `filename` as dictionary.
        """
        self.dirty = False
        self.template = None
        with open(self.filename, 'r') as f:
            if self.template_format == '.json':
                self.template = json.load(f, object_pairs_hook=collections.OrderedDict)
            else:
                yaml = YAML()
                self.template = yaml.load(f)

    def is_cloudformation_template(self):
        """
        returns true if the `self.template` is a AWS CloudFormation template
        """
        return 'AWSTemplateFormatVersion' in self.template

    def write(self):
        """
        write modified content from `template` to `filename`. It will retain it's original
        format (yaml or json) but loose original formatting and comments.
        """
        if not self.dirty:
            if self.verbose:
                sys.stderr.write('INFO: no changes in {}\n'.format(self.filename))
            return

        if self.dry_run:
            return

        with open(self.filename, 'w') as f:
            if self.template_format == '.yaml':
                yaml = YAML()
                yaml.dump(self.template, f)
            else:
                json.dump(self.template, f, separators=(',', ': '), indent=2)

    def update_template(self):
        """
        implement the update logic of `self.template`. Set self.dirty to True, if you modified.
        """
        pass

    @property
    def resources(self):
        return self.template.get('Resources', {})

    def update(self, path):
        """
        recursively updates all the cloudformation templates in the specified `path`. `path` may be a file,
        a directory or a list of paths.
        """
        if isinstance(path, list):
            for p in path:
                self.update(p)
        elif os.path.isfile(path):
            if path.endswith('.yml') or path.endswith('.yaml') or path.endswith('.json'):
                self.filename = path
                self.load()
                if self.is_cloudformation_template():
                    self.update_template()
                    self.write()
                else:
                    if self.verbose:
                        sys.stderr.write('INFO: skipping {} as it is not a CloudFormation template\n'.format(path))
        elif os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for f in files:
                    self.update(os.path.join(root, f))
        else:
            sys.stderr.write('ERROR: {} is not a file or directory\n'.format(path))
            sys.exit(1)


class Ref:
    yaml_tag = u'!Ref'

    def __init__(self, reference):
        self.reference = reference

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.reference)

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.value)

yaml = YAML()
yaml.register_class(Ref)


#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import re
import sys
from datetime import datetime, timedelta, tzinfo

from croniter import croniter
from tzlocal import get_localzone

from .cfn_updater import CfnUpdater


class CronScheduleExpressionUpdater(CfnUpdater):
    """
    Updates the schedule expression of an AWS::Events::Rules resources to reflect the
    scheduled time in UTC. The required cron rule is taken from the description. It will
    update the following resource definition from:

\b
      DailyTaskSchedule:
        Type: AWS::Events::Rule
        Properties:
          Description: run daily - cron(30 01 * * ? *)
          Name: run daily
          ScheduleExpression: cron(30 01 * * ? *)
          State: ENABLED

to

\b
      DailyTaskSchedule:
        Type: AWS::Events::Rule
        Properties:
          Description: run daily - cron(30 01 * * ? *)
          Name: run daily
          ScheduleExpression: cron(30 23 * * ? *)
          State: ENABLED

with --timezone Europe/Amsterdam and --date 2018-08-01. If the updater is run with --date 2018-12-01, it
will change it to:

\b
      DailyTaskSchedule:
        Type: AWS::Events::Rule
        Properties:
          Description: run daily - cron(30 01 * * ? *)
          Name: run daily
          ScheduleExpression: cron(30 00 * * ? *)
          State: ENABLED


    """

    def __init__(self):
        super(CronScheduleExpressionUpdater, self).__init__()
        self._timezone = get_localzone()

        self._today = datetime.now()
        self.verbose = False
        self.dry_run = False

    @property
    def timezone(self):
        return self._timezone

    @timezone.setter
    def timezone(self, timezone):
        self._timezone = timezone
        self._today = timezone.localize(self._today)

    @property
    def today(self):
        return self._today

    @today.setter
    def today(self, dt):
        if dt.tzinfo:
            self._today = dt.replace(tzinfo=self.timezone)
        else:
            self._today = self.timezone.localize(dt)

    def is_matching_resource(self, resource):
        aws_type = resource.get('Type', '')
        description = resource.get('Properties', {}).get('Description', '')
        expression = resource.get('Properties', {}).get('ScheduleExpression', None)

        if aws_type == 'AWS::Events::Rule' and aws_cron_pattern.search(description):
            return expression is None or aws_cron_pattern.search(expression)

        return False

    def all_matching_resources(self, resources):
        return {k: v for k, v in resources.items() if self.is_matching_resource(v)}

    def update_template(self):
        """
        converts the cron expression in `Description` into a UTC expression in `ScheduleExpression`.
        """
        resources = self.all_matching_resources(self.template.get('Resources', {}))

        for name, resource in resources.items():
            description = resource.get('Properties', {}).get('Description', '')
            match = aws_cron_pattern.search(description)
            if match:
                expression = '{minutes} {hours} {day_of_month} {month} {day_of_week} {year}'.format(**match.groupdict())
                new_expression = correct_cron_expression_for_utc(expression, self.today)
                if expression != new_expression:
                    properties = resource.get('Properties')
                    properties['ScheduleExpression'] = 'cron({})'.format(new_expression)
                    if self.verbose:
                        print('INFO: updating {}'.format(name))
                    self.dirty = True

    def main(self, tz, date, dry_run, verbose, paths):
        self.dry_run = dry_run
        self.verbose = verbose
        self.timezone = tz
        if date:
            self.today = date
        self.update(paths)


aws_cron_pattern = re.compile(
    r'cron\s*\(\s*(?P<minutes>[^\s]*)\s*(?P<hours>[^\s]*)\s*(?P<day_of_month>[^\s]*)\s*(?P<month>[^\s]*)\s*(?P<day_of_week>[^\s]*)\s*(?P<year>[^\s\)]*)\)')
cron_pattern = re.compile(
    r'\s*(?P<minutes>[^\s]*)\s*(?P<hours>[^\s]*)\s*(?P<day_of_month>[^\s]*)\s*(?P<month>[^\s]*)\s*(?P<day_of_week>[^\s]*)\s*(?P<year>[^\s\)]*)')


def correct_for_utc(hour, utcoffset):
    if utcoffset.seconds % 3600 == 0:
        return int((hour + 24 - (utcoffset.seconds / 3600)) % 24)
    else:
        raise ValueError('UTC offset is not a multiple of hours.')


def correct_cron_hours_expression_for_utc(expression, utcoffset):
    if expression == '*' or expression == '?':
        return expression

    range = re.match(r'([0-9]+)-([0-9]+)', expression)
    if range:
        from_hour = int(range.group(1))
        to_hour = int(range.group(2))
        return '{}-{}'.format(correct_for_utc(from_hour, utcoffset), correct_for_utc(to_hour, utcoffset))

    repeat = re.match(r'([0-9]+)/([0-9]+)', expression)
    if repeat:
        from_hour = int(repeat.group(1))
        repetition = int(repeat.group(2))
        return '{}/{}'.format(correct_for_utc(from_hour, utcoffset), repetition)

    hours = list(filter(lambda n: re.match(r'[0-9]+', n), expression.split(',')))
    if len(hours) == len(expression.split(',')):
        hours = list(map(lambda h: '{}'.format(correct_for_utc(int(h), utcoffset)), hours))
        return ','.join(hours)
    else:
        assert False, 'unsupported hour format {}'.format(expression)

    return expression


def correct_cron_expression_for_utc(expression, today):
    match = cron_pattern.search(expression)
    assert match, '"{}" is not a cron expression'.format(expression)

    cron = match.groupdict()
    tomorrow_midnight = today.tzinfo.localize(datetime(today.year, today.month, today.day) + timedelta(days=1))

    
    try:
        ccron = {k:('*' if v == '?' else v) for (k,v) in cron.items()}
        expression = '{minutes} {hours} {day_of_month} {month} {day_of_week} {year}'.format(**ccron)
        next_time = croniter(expression, tomorrow_midnight).get_next(datetime)
    except ValueError as e:
        sys.stderr.write('ERROR: {}'.format(e))
        sys.exit(1)

    utcoffset = next_time.tzinfo.utcoffset(next_time)
    if utcoffset.seconds % 3600 == 0:
        cron['hours'] = correct_cron_hours_expression_for_utc(cron['hours'], utcoffset)
        expression = '{minutes} {hours} {day_of_month} {month} {day_of_week} {year}'.format(**cron)
    else:
        sys.stderr.write(
            'WARN: UTC offset for "{}" in timezone "{}" is not a multiple of hours but {} seconds.\n'.format(expression,
                                                                                                             today.tzinfo,
                                                                                                             utcoffset.seconds))

    return expression



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from .cfn_updater import CfnUpdater


class ContainerImageUpdater(CfnUpdater):
    """
    Updates the Docker image of ECS Container Definitions.

    it will update any container definition where the base image name matches
    the specified image name excluding the tag.

    For example, an image name of `mvanholsteijn/paas-monitor:0.6.0` will update:

\b
        Type: AWS::ECS::TaskDefinition
        Properties:
          ContainerDefinitions:
            - Name: paas-monitor
              Image: mvanholsteijn/paas-monitor:0.5.9

    to:

\b
        Type: AWS::ECS::TaskDefinition
        Properties:
          ContainerDefinitions:
            - Name: paas-monitor
              Image: mvanholsteijn/paas-monitor:0.6.0

    """

    def __init__(self):
        super(ContainerImageUpdater, self).__init__()
        self._image = []

    @property
    def image(self):
        return ':'.join(self._image)

    @image.setter
    def image(self, image):
        self._image = image.split(':', 2) if image is not None else []

    @property
    def base_image(self):
        return self._image[0] if self._image is not None and len(self._image) > 0 else ''

    @property
    def image_tag(self):
        return self._image[1] if self._image is not None and len(self._image) > 1 else ''

    @staticmethod
    def is_task_definition(resource):
        """
        returns true if the resource is of type AWS::ECS::TaskDefinition
        """
        return resource.get('Type', '') == 'AWS::ECS::TaskDefinition'

    def is_matching_container(self, container):
        """
        returns true if there is a match on the image
        """
        return container.get('Image', '').split(':')[0] == self.base_image

    def all_matching_task_definitions(self, resources):
        return filter(lambda n: self.is_task_definition(resources[n]), resources)

    def update_template(self):
        """
        updates the Image of container definitions in the CFN template `self.template`.
        """
        resources = self.template.get('Resources', {})
        for task_name in self.all_matching_task_definitions(resources):
            task = resources[task_name]
            containers = task.get('Properties', {}).get('ContainerDefinitions', [])
            for container in filter(lambda c: self.is_matching_container(c), containers):
                if container['Image'] != self.image:
                    sys.stderr.write(
                        'INFO: updating image of container definition {} for task {} in {}\n'.format(
                            container['Name'], task_name, self.filename))
                    container['Image'] = self.image
                    self.dirty = True

    def main(self, image, dry_run, verbose, paths):
        self.image = image
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)
